#!/usr/bin/env python3

import argparse
import warnings
import datetime as dt
from timeit import Timer

from lbmpy.session import *
from pystencils.field import create_numpy_array_with_layout
from pystencils.kerncraft_coupling import *
from pystencils.sympyextensions import prod
from pystencils.timeloop import TimeLoop
from kerncraft.machinemodel import MachineModel
from kerncraft.models import ECM, Benchmark, ECMData
from kerncraft.prefixedunit import PrefixedUnit

sep = '-----------------------------------------\n'
results_filepath = ''
MACHINE_FILE_PATH = './machine-files/'
global_arch = ''


def main():
    parser = create_parser()
    args = parser.parse_args()
    check_arguments(args, parser)
    cse_g, cse_pdfs, vec, comp, split, dims = init_parameters(args)
    run_benchmarks(
        cse_g, cse_pdfs, vec, comp, split, dims, args.openmp, args.with_pheno
    )


def create_parser():
    # Create parser
    parser = argparse.ArgumentParser(
        description='Runs lbmpy benchmarks and returns results in a plottable format.'
    )

    # Add arguments
    parser.add_argument(
        'arch', type=str, help='Target architecture. Must be one of [SKL, HSW, IVB].'
    )
    parser.add_argument(
        '-o',
        '--output',
        dest='output',
        type=str,
        help=(
            'Destination output file. If no parameter is given, the output is written to with the '
            'current timestamp to "results_ARCH.TIMESTAMP.csv".'
        )
    )
    parser.add_argument(
        '--all', action='store_true', help='Run benchmarks with all possible options.'
    )
    parser.add_argument(
        '-f',
        '--fastest',
        action='store_true',
        dest='fastest',
        help='Run only fastest benchmark options.',
    )
    parser.add_argument(
        '-d',
        '--dims',
        action='store_true',
        dest='dims',
        help=(
            'Run a dimension scaling for x in the range [10, 25, 50, 100, 125, 150, 175, 200, 250,'
            ' 300, 400, 500]. Otherwise, a fixed x dimension of 300 is assumed.'
        ),
    )
    parser.add_argument(
        '--openmp', action='store_true', help='Creates a parallel version of the code with OpenMP.'
    )
    parser.add_argument(
        '--with-pheno',
        action='store_true',
        help=('Applies a phenomenological ECM model for Kerncraft analysis.'),
    )
    return parser


def check_arguments(args, parser):
    global results_filepath
    global global_arch

    supported_archs = ['IVB', 'HSW', 'SKL']
    if args.arch.upper() not in supported_archs:
        parser.error(
            'Microarchitecture not supported. Please see --help for all valid architecture codes.'
        )
    global_arch = args.arch.lower()
    ts = dt.datetime.utcnow().timestamp()
    results_filepath = (
        args.output if args.output is not None else 'results_{}.{}.csv'.format(args.arch.lower(), ts)
    )


def init_parameters(args):
    arch = args.arch.upper()
    if args.all:
        cse_g = [True, False]
        cse_pdfs = [True, False]
        vect = [
            {'instruction_set': 'avx', 'assume_aligned': True, 'nontemporal': True},
            {'instruction_set': 'avx', 'assume_aligned': True, 'nontemporal': False},
            {'instruction_set': 'avx', 'assume_aligned': False, 'nontemporal': False},
        ]
        if arch == 'SKL':
            vect = [
                {'instruction_set': 'avx512', 'assume_aligned': True, 'nontemporal': True},
                {'instruction_set': 'avx512', 'assume_aligned': True, 'nontemporal': False},
                {'instruction_set': 'avx512', 'assume_aligned': False, 'nontemporal': False},
            ]
        comp = [True, False]
        split = [True, False]
    elif args.fastest:
        if arch == 'SKL':
            cse_g = split = [True]
            comp = cse_pdfs = [False]
            vect = [{'instruction_set': 'avx512', 'assume_aligned': True, 'nontemporal': False}]
        elif arch == 'HSW':
            cse_g = comp = [False]
            cse_pdfs = split = [True]
            vect = [{'instruction_set': 'avx', 'assume_aligned': True, 'nontemporal': True}]
        elif arch == 'IVB':
            cse_g = comp = [False]
            cse_pdfs = split = [True]
            vect = [{'instruction_set': 'avx', 'assume_aligned': True, 'nontemporal': False}]
    dims = [10, 25, 50, 100, 125, 150, 175, 200, 250, 300, 400, 500] if args.dims else [300]
    return cse_g, cse_pdfs, vect, comp, split, dims


#######################################
# MAIN FUNCTIONS


def run_benchmarks(cse_g, cse_pdfs, vec, comp, split, dims, openmp, with_pheno):
    cse_global_options = cse_g
    cse_pdfs_options = cse_pdfs
    vect_options = vec
    compress_options = comp
    split_options = split
    dims_x = dims
    total = (
        len(cse_global_options)
        * len(cse_pdfs_options)
        * len(vect_options)
        * len(compress_options)
        * len(split_options)
        * len(dims_x)
    )
    curr = 1
    for cgo in cse_global_options:
        for cpo in cse_pdfs_options:
            for vo in vect_options:
                for co in compress_options:
                    for so in split_options:
                        for dim_x in dims_x:
                            shape = [dim_x, 100, 100, 19]
                            opt = {
                                'cse_pdfs': cpo,
                                'cse_global': cgo,
                                'split': so,
                                'vectorization': vo,
                                'openmp': openmp,
                            }
                            m_opt = {
                                'method': 'trt',
                                'stencil': 'D3Q19',
                                'relaxation_rates': [1.6, 1.8],
                                'compressible': co,
                            }
                            print('{}/{}'.format(curr, total))
                            print(get_param_info(m_opt, opt, shape))
                            if curr == 1:
                                write_csv_header(
                                    m_opt['method'],
                                    m_opt['stencil'],
                                    m_opt['relaxation_rates'],
                                    shape,
                                )
                            kerncraft_bench(
                                m_opt, opt, shape, loop='all', with_pheno=with_pheno
                            )
                            print(sep + sep, end='')
                            curr += 1


def init(m_opt=None, opts=None, shape=None):
    shape = shape if shape is not None else [100, 100, 100, 19]
    if m_opt is None:
        m_opt = dict()
    if (
        opts is not None
        and 'assume_aligned' in opts['vectorization']
        and opts['vectorization']['assume_aligned']
    ):
        # with assume_aligned
        src_arr = create_numpy_array_with_layout(
            shape=shape, layout=(3, 2, 1, 0), alignment=8 * 8, byte_offset=8
        )
    else:
        # without assume_aligned
        src_arr = np.zeros(shape, order='f')
    print("Src array strides ", [a // 8 for a in src_arr.strides])
    dst_arr = np.zeros_like(src_arr)
    cells = prod(t - 2 for t in src_arr.shape[:-1])  # subtract ghost layers
    opt = {'cse_pdfs': True, 'split': False, 'vectorization': {'instruction_set': 'avx'}}
    opt = opts if opts is not None else opt
    opt['openmp'] = False if 'openmp' not in opt else opt['openmp']
    opt['symbolic_field'] = ps.Field.create_from_numpy_array('src', src_arr, index_dimensions=1)
    method = 'trt' if 'method' not in m_opt else m_opt['method']
    stencil = 'D3Q19' if 'stencil' not in m_opt else m_opt['stencil']
    compr = False if 'compressible' not in m_opt else m_opt['compressible']
    relaxation = [1.8, 1.6] if 'relaxation_rates' not in m_opt else m_opt['relaxation_rates']
    ast = create_lb_ast(
        #        method='trt', stencil='D3Q19', optimization=opt, relaxation_rates=[1.8, 1.6]
        method=method,
        stencil=stencil,
        optimization=opt.copy(),
        compressible=compr,
        relaxation_rates=relaxation,
    )
    used_opts = opt.copy()
    used_opts['compressible'] = compr
    return ast, cells, used_opts


def c_bench(m_opt=None, opt=None, shape=None):
    ast, cells, used_opts = init(m_opt, opt, shape)
    print('CELLS: {}'.format(cells))

    print('{}C benchmark'.format(sep))
    timings = run_c_benchmark(
        ast,
        inner_iterations=20,
        outer_iterations=5,
        assume_aligned=opt['vectorization']['assume_aligned'],
        # openmp=opt['openmp']
    )
    mlups = cells / min(timings) * 1e-6
    print('MFLUPs: {}, ({}s)'.format(mlups, timings))
    # write_c_bench_val(str(round(mlups,2))+' MFLUPs')
    return mlups


def kerncraft_bench(m_opt=None, opt=None, shape=None, loop=False, with_pheno=False):
    ast, cells, used_opts = init(m_opt, opt, shape)
    # print(show_code(ast))
    print('{}Kerncraft'.format(sep))
    machine_file_path = MACHINE_FILE_PATH
    if global_arch == 'skl':
        machine_file_path += 'SkylakeSP_Gold-6148.yml'
    elif global_arch == 'hsw':
        machine_file_path += 'HaswellEP_E5-2695v3_CoD.yml'
    elif global_arch == 'ivb':
        machine_file_path += 'IvyBridgeEP_E5-2660v2.yml'
    machine = MachineModel(path_to_yaml=machine_file_path)
    loops = PyStencilsKerncraftKernel.get_number_of_kernels(ast)
    steps = [s / 8 for s in PyStencilsKerncraftKernel.get_steps_for_loops(ast)]
    results = []
    print('Loops: {}'.format(loops))
    if loop:
        if loop == 'all':
            for l in range(loops):
                print('Loop {}'.format(l))
                results.append(
                    kerncraft_helper(
                        ast=ast,
                        machine=machine,
                        loop_idx=l,
                        accumulate=True,
                        with_pheno=with_pheno,
                        debug_print=False,
                    )
                )
                print('')
            for i, r in enumerate(results):
                print(
                    'Loop {} ({} steps): {}, FLOPS: {}'.format(
                        i, int(steps[i]), r[0]['cy/CL'], r[1]._flops
                    )
                )
            p_sum = get_loop_sum(results, machine)
            print('Sum: {}'.format(p_sum))
            all_cy_cl = sum_prefixedUnits([r[0]['cy/CL'] for r in results])
            arch = global_arch.upper()
            write_csv_row(used_opts, arch, shape[0], p_sum, all_cy_cl)
        elif loops > loops:
            warnings.warn('chosen loop is out of pystencils AST loop range - choosing first one')
            kerncraft_helper(
                ast=ast, machine=machine, loop_idx=1, with_pheno=with_pheno, debug_print=False
            )
        else:
            kerncraft_helper(
                ast=ast, machine=machine, loop_idx=loop, with_pheno=with_pheno, debug_print=False
            )
    c_mlups = c_bench(m_opt, opt, shape)
    c_cycl = conv_to_cycl(c_mlups, results[0][1].datatypes_size[results[0][1].datatype], machine)
    write_c_bench_val(str(round(c_mlups, 2)), str(round(c_cycl.value, 2)))


#######################################
# HELPER FUNCTIONS


def get_param_info(m_opt, opt, shape):
    inf = (
        'Method parameters\n'
        + '-----------------\n'
        + 'method={}\n'
        + 'stencil={}\n'
        + 'relaxation_rates={}\n'
        + 'compressible={}\n\n'
        + 'Optimization parameters\n'
        + '-----------------------\n'
        + 'cse_pdfs={}\n'
        + 'cse_global={}\n'
        + 'split={}\n'
        + 'vectorization={}\n'
        + 'shape={}\n'
    ).format(
        m_opt['method'],
        m_opt['stencil'],
        m_opt['relaxation_rates'],
        m_opt['compressible'],
        opt['cse_pdfs'],
        opt['cse_global'],
        opt['split'],
        opt['vectorization'],
        shape,
    )
    return inf


def kerncraft_helper(ast, machine, loop_idx, accumulate=False, with_pheno=True, debug_print=False):
    kc_kernel = PyStencilsKerncraftKernel(
        ast=ast, machine=machine, loop_idx=loop_idx, debug_print=debug_print
    )
    print('FLOPS: {}'.format(kc_kernel._flops))
    steps = PyStencilsKerncraftKernel.get_steps_for_loops(ast)
    if with_pheno:
        kernbench = Benchmark(kc_kernel, machine, KerncraftParameters(ptr_inc=steps[loop_idx]))
        kernbench.analyze()
        print(kernbench.report())
    ecm = ECM(kc_kernel, machine, KerncraftParameters(ptr_inc=steps[loop_idx]))
    ecm.analyze()
    print(ecm.report())
    if accumulate:
        # get single core performance prediction
        return ecm.results.get('scaling prediction')[0]['performance'], kc_kernel


def kerncraft_pheno_helper(ast, machine, loop_idx, debug_print):
    kc_kernel = PyStencilsKerncraftKernel(
        ast=ast, machine=machine, loop_idx=loop_idx, debug_print=debug_print
    )
    print('FLOPS: {}'.format(kc_kernel._flops))
    steps = PyStencilsKerncraftKernel.get_steps_for_loops(ast)
    kernbench = Benchmark(kc_kernel, machine, KerncraftParameters(ptr_inc=steps[loop_idx]))
    kernbench.analyze()
    print(kernbench.report())


def get_loop_sum(results, machine):
    all_cy_cl = sum_prefixedUnits([r[0]['cy/CL'] for r in results])
    element_size = results[0][1].datatypes_size[results[0][1].datatype]
    elements_per_cacheline = int(machine['cacheline size']) // element_size
    it_s = machine['clock'] / all_cy_cl * elements_per_cacheline
    it_s.unit = 'It/s'
    return it_s


def conv_to_cycl(mlups, element_size, machine):
    clock = machine['clock']
    elem_per_cl = int(machine['cacheline size']) // element_size
    cy_cl = clock / PrefixedUnit(mlups * 1000000, '', u'It/s') * elem_per_cl
    cy_cl.unit = 'cy/CL'
    return cy_cl


def sum_prefixedUnits(pu_list):
    res = pu_list[0]
    for pu in pu_list[1:]:
        res += pu
    return res


def write_csv_row(opts, arch, dim_x, p_ecm, ecm_cyCL):
    row = '{},{},{},{},{},{},{},{},{},{},{},'.format(
        opts['compressible'],
        opts['cse_pdfs'],
        opts['cse_global'],
        opts['split'],
        opts['vectorization']['instruction_set'],
        opts['vectorization']['assume_aligned'],
        opts['vectorization']['nontemporal'],
        arch,
        dim_x,
        round((p_ecm * 1e-6).value, 2)
        if p_ecm.prefix == ''
        else round(p_ecm.value, 2) + p_ecm.prefix + p_ecm.unit,
        round(ecm_cyCL.value, 2)
        if ecm_cyCL.prefix == ''
        else round(ecm_cyCL.value, 2) + ecm_cyCL.prefix + ecm_cyCL.unit,
    )
    with open(results_filepath, 'a') as f:
        f.write(row)


def write_c_bench_val(p_mflups, p_cycl):
    with open(results_filepath, 'a') as f:
        f.write('{},{}\n'.format(p_mflups, p_cycl))


def write_csv_header(method, stencil, relaxation, shape):
    comment_line = '#method={}, stencil={}, relaxation_rates={}, shape=[dim_x,{},{}]\n'.format(
        method, stencil, relaxation, shape[1], shape[2]
    )
    header = (
        'compressible,cse_pdfs,cse_global,split,vec_opt,aligned,nontemp,'
        'arch,dim_x,ECM[MIt/s],ECM_cyCL[cy/CL],c_bench[MFLUPs],c_bench_cyCL[cy/CL]\n'
    )
    with open(results_filepath, 'a') as f:
        f.write(comment_line + header)


#######################################

if __name__ == '__main__':
    main()
